(use gauche.test)

(test-start "text.yaml")
(use text.yaml)
(test-module 'text.yaml)

(test-section "boundary case")

(test* "parsing a null string yields the eof object." #t
       (eof-object? (yaml-parse-string "")))

(test-section "boolean")

(test* "Boolean(yes)" #t (yaml-parse-string "yes"))

(test* "Boolean(true)" #t (yaml-parse-string "true"))

(test* "Boolean(on)" #t (yaml-parse-string "on"))

(test* "Boolean(no)" #f (yaml-parse-string "no"))

(test* "Boolean(false)" #f (yaml-parse-string "false"))

(test* "Boolean(off)" #f (yaml-parse-string "off"))


(test-section "null")

(test* "null" (undefined) (yaml-parse-string "null"))

(test* "null(~)" (undefined) (yaml-parse-string "~"))


(test-section "integer")

(test* "decimal integer" 685230 (yaml-parse-string "685230"))

;; Syck uses `,' as a separator instead of `_', which is specified in http://yaml.org/type/int.html
(test* "decimal integer(with plus sign and comma)" 685230
       (yaml-parse-string "+685,230"))

(test* "octal integer" 685230 (yaml-parse-string "02472256"))

(test* "hexadecimal integer(mixed case, comma separated)" 685230
       (yaml-parse-string "0x,0A,74,AE"))

;; Syck does not support binary integer.
;;(test* "binary integer(comma separated)" 685230
;;       (yaml-parse-string "0b1010,0111,0100,1010,1110"))

(test* "sexagesimal integer" 685230 (yaml-parse-string "190:20:30"))


(test-section "float")

(test* "float(canonical)" 685230.15 (yaml-parse-string "6.8523015e+5"))

(test* "float(exponential)" 685230.15 (yaml-parse-string "685.230.15e+03"))

(test* "float(exponential)" 685230.15 (yaml-parse-string "685.23015e+03"))

(test* "float(fixed, with comma)" 685230.15 (yaml-parse-string "685,230.15"))

(test* "float(sexagesimal)" 685230.15 (yaml-parse-string "190:20:30.15"))

(test* "float(infinity)" (/ 1.0 0.0) (yaml-parse-string ".inf"))

(test* "float(negative infinity)" (/ -1.0 0.0) (yaml-parse-string "-.inf"))

(test* "float(nan)" (/ 0.0 0.0) (yaml-parse-string ".nan"))


(test-section "string")

(test* "string" "abcd" (yaml-parse-string "abcd"))


(test-section "time stamp")

(use srfi-19)

(define-method object-equal? ((d1 <date>) (d2 <date>))
  (zero? (sys-difftime (date->time-utc d1) (date->time-utc d2))))

(let ((d1 (make-date 100000000 43 59 2 15 12 2001 0))
      (d2 (make-date 0 0 0 0 14 12 2002 0)))
  (test* "date (canonical)" d1
	 (yaml-parse-string "2001-12-15T02:59:43.1Z"))
  
  (test* "date (iso8601)" d1
	 (yaml-parse-string "2001-12-14t21:59:43.10-05:00"))
  
   (test* "date (space separated)" d1
;; Syck assumes that hour, etc. should be two digits.
;; 	 (yaml-parse-string "2001-12-14 21:59:43.10 -5")
 	 (yaml-parse-string "2001-12-14 21:59:43.10 -05"))

   (test* "date (no timezone)" d1
;; Syck does not support space separated timestamps without timezone `Z'.
;; 	 (yaml-parse-string "2001-12-15 2:59:43.10")
 	 (yaml-parse-string "2001-12-15 02:59:43.10 Z"))

   (test* "date (yyyy-mm-dd)" d2
	  (yaml-parse-string "2002-12-14"))
  )

(use util.match)

(for-each
 (match-lambda
  ((num . str)
   (test* "date (fractional part. extra numbers shold be dropped)" num
	  (date-nanosecond
	   (yaml-parse-string (format "2002-12-15 02:59:43~A Z" str))))))
 '((000000000 . "")
   (100000000 . ".1")
   (110000000 . ".11")
   (111000000 . ".111")
   (111100000 . ".1111")
   (111110000 . ".11111")
   (111111000 . ".111111")
   (111111100 . ".1111111")
   (111111110 . ".11111111")
   (111111111 . ".111111111")
   (111111111 . ".1111111111")
   (111111111 . ".11111111111")))


(test-section "sequences")

(let1 xs '("Mercury" "Venus" "Earth" "Mars"
	   "Jupiter" "Saturn" "Uranus" "Neptune"
	   "Pluto")
  (test* "block-style" xs
	 (yaml-parse-string "
- Mercury   # Rotates - no light/dark sides.
- Venus     # Deadliest. Aptly named.
- Earth     # Mostly dirt.
- Mars      # Seems empty.
- Jupiter   # The king.
- Saturn    # Pretty.
- Uranus    # Where the sun hardly shines.
- Neptune   # Boring. No rings.
- Pluto     # You call this a planet?"))

  (test* "flow style" xs
	 (yaml-parse-string "
[ Mercury, Venus, Earth, Mars,      # Rocks
  Jupiter, Saturn, Uranus, Neptune, # Gas
  Pluto ]"))
  )


(test-section "map")

(use srfi-1)
(use util.list)

(let1 alis '(("Clark" . "Evans")
	     ("Brian" . "Ingerson")
	     ("Oren"  . "Ben-Kiki"))
  (define (pred alist ht)
    (and (= (length alist) (hash-table-num-entries ht))
	 (every (lambda (pair)
		  (and (hash-table-exists? ht (car pair))
		       (equal? (ref ht (car pair)) (cdr pair))))
		alist)))
  (test* "block style" alis
	 (yaml-parse-string "
Clark : Evans
Brian : Ingerson
Oren  : Ben-Kiki")
	 pred)

  (test* "flow style" alis
	 (yaml-parse-string
	  "{ Clark: Evans, Brian: Ingerson, Oren: Ben-Kiki }")
	 pred)
  )

(test* "ignore !!value" 0
       (hash-table-num-entries (yaml-parse-string "= : a")))

(test* "ignore !!value" 1
       (hash-table-num-entries (yaml-parse-string "= : a\na : b")))


(test-section "read/write invariance")

(for-each
 (lambda (obj)
   (test* (write-to-string obj) obj
	  (yaml-parse-string (x->yaml obj))))
 (list #t #f 0 #x14 '() #i0/0 #i1/0 #i-1/0 "true"
       (current-date) '("abc" ("def" ("ghi" "jkl")))))


;; hash-table cannot be compare with equal?
;; and object-equal? for built-in class cannot be redefine.
(define (hash-table-equal? ht1 ht2)
  (and (hash-table? ht1) (hash-table? ht2)
       (= (hash-table-num-entries ht1) (hash-table-num-entries ht2))
       (every (lambda (key)
		(and (hash-table-exists? ht2 key)
		     (or (equal? (ref ht1 key) (ref ht2 key))
			 (hash-table-equal? (ref ht1 key) (ref ht2 key)))))
	      (hash-table-keys ht1))))

(for-each
 (lambda (ht)
   (test* (write-to-string ht) ht
	  (yaml-parse-string (x->yaml ht))
	  hash-table-equal?))
 (list (make-hash-table 'equal?)
       (hash-table 'equal?
		   '("abc" . "def") '(1 . 2) '(3 . (4 5 6))
		   (cons "today" (list (current-date) (current-date))))
       (hash-table 'equal?
		   '(("abc" "def") . "ghi")
		   '((1 2) . (3 4)))))

(let1 ht (hash-table 'equal?
		     '(() . 1)
		     '(() . 2))
  (test* "empty seq for map key" ht
	 (yaml-parse-string (x->yaml ht))
	 hash-table-equal?))

(let1 ht (hash-table 'equal?
		     '(1 . ())
		     (cons 2 (make-hash-table 'equal?)))
  (test* "empty seq and empty map for map value" ht
	 (yaml-parse-string (x->yaml ht))
	 hash-table-equal?))


(test-section "simple strings")

(for-each
 (lambda (str)
   (test* str str (yaml-parse-string (x->yaml str))))
 '(".inf" ".Inf" "+.inf" "-.inf" ".nan" ".NaN" "null" "Null"
   "yes" "Yes" "y" "Y" "NO" "no" "true" "False" "on" "OFF"
   "2001-12-14 21:59:43.10 -05"
   ))


(test-section "binary")

(use gauche.uvector)
(use rfc.base64)

(let1 b64data "R0lGODlhDAAMAIQAAP//9/X17unp5WZmZgAAAOfn515eXvPz7Y6OjuDg4J+fn5OTk6enp56enmlpaWNjY6Ojo4SEhP/++f/++f/++f/++f/++f/++f/++f/++f/+ +f/++f/++f/++f/++f/++SH+Dk1hZGUgd2l0aCBHSU1QACwAAAAADAAMAAAFLCAgjoEwnuNAFOhpEMTRiggcz4BNJHrv/zCFcLiwMWYNG84BwwEeECcgggoBADs="
  (test* "base64 decode" (base64-decode-string b64data)
	 (u8vector->string
	  (yaml-parse-string (string-append "!binary |\n" b64data))))
  (test* "base64 encode" (base64-decode-string b64data)
	 (u8vector->string
	  (yaml-parse-string
	   (x->yaml
	    (string->u8vector (base64-decode-string b64data))))))
  )

(let1 obj '(#u8(0 1 2) #u8(3 4 5) (#u8(6 7 8) #u8(9 10)))
  (yaml-dump obj)
  (test* "seq of binary" obj (yaml-parse-string (x->yaml obj)))) 


(test-section "merge")

(test* "merging" #t
       (yaml-parse-string "
- &CENTER { x: 1, y: 2 }
- &LEFT { x: 0, y: 2 }
- &BIG { r: 10 }
- &SMALL { r: 1 }

# All the following maps are equal:

- # Explicit keys
  x: 1
  y: 2
  r: 10
  label: center/big

- # Merge one map
  << : *CENTER
  r: 10
  label: center/big

- # Merge multiple maps
  << : [ *CENTER, *BIG ]
  label: center/big

- # Override
  << : [ *BIG, *LEFT, *SMALL ]
  x: 1
  label: center/big")
       (lambda (_ xs)
	 (and
	  (hash-table-equal? (fifth xs) (sixth xs))
	  (hash-table-equal? (sixth xs) (seventh xs))
	  (hash-table-equal? (seventh xs) (eighth xs))
	  #t)))


(test-end)
