;;;
;;; Copyright (c) 2006 OOHASHI Daichi, All rights reserved.
;;;
;;; Redistribution and use in source and binary forms, with or without
;;; modification, are permitted provided that the following conditions
;;; are met:
;;;
;;; 1. Redistributions of source code must retain the above copyright
;;;    notice, this list of conditions and the following disclaimer.
;;;
;;; 2. Redistributions in binary form must reproduce the above copyright
;;;    notice, this list of conditions and the following disclaimer in the
;;;    documentation and/or other materials provided with the distribution.
;;;
;;; 3. Neither the name of the authors nor the names of its contributors
;;;    may be used to endorse or promote products derived from this
;;;    software without specific prior written permission.
;;;
;;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;;; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;;; A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
;;; OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;;; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
;;; TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
;;; PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
;;; LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
;;; NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
;;; SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;;;

(define-module text.yaml
  (use srfi-19)
  (use gauche.parameter)
  (use gauche.uvector)
  (use rfc.base64)
  (export yaml-parse-string yaml-load yaml-dump x->yaml)
  )

(select-module text.yaml)
(dynamic-load "yaml")

(define (yaml-load port)
  (yaml-parse-string (port->string port)))


(define (x->yaml obj)
  (call-with-output-string (cut yaml-dump obj <>)))


(define-method yaml-dump (obj)
  (yaml-dump obj (current-output-port)))


;; utilities
(define current-indent (make-parameter 0))

(define (indent str)
  (string-append (make-string (current-indent) #\space) str))

(define-syntax with-indent
  (syntax-rules ()
    ((_ amount body ...)
     (parameterize ((current-indent (+ (current-indent) amount)))
       body ...))))

(define (yaml-collection? obj)
  (or (null? obj) (pair? obj) (hash-table? obj)))

(define (yaml-empty-collection? obj)
  (or (null? obj)
      (and (hash-table? obj)
	   (zero? (hash-table-num-entries obj)))))

(define (yaml-binary? obj)
  (or (string-incomplete? obj) (u8vector? obj)))

(define (yaml-multiline? obj)
  (or (and (yaml-collection? obj)
	   (not (yaml-empty-collection? obj)))
      (yaml-binary? obj)))


;;; Collection types
;; !!seq
(define-method yaml-dump ((null <null>) port)
  (display "[]" port))

(define-method yaml-dump ((xs <list>) port)
  (for-each
   (lambda (x)
     (display (indent "- ") port)
     (cond ((yaml-multiline? x)
	    (newline port)
	    (with-indent 2 (yaml-dump x port)))
	   (else
	    (yaml-dump x port)
	    (newline port))))
   xs))

;; !!map
(define-method yaml-dump ((ht <hash-table>) port)
  (cond
   ((zero? (hash-table-num-entries ht))
    (display "{}" port))
   (else
    (hash-table-for-each ht
      (lambda (key val)
	(cond ((yaml-empty-collection? key)
	       (display (indent "? ") port)
	       (yaml-dump key port)
	       (newline port))
	      ((yaml-multiline? key)
	       (display (indent "? ") port)
	       (newline port)
	       (with-indent 2 (yaml-dump key port)))
	      (else
	       (display (indent "") port)
	       (yaml-dump key port)))
	(display ": " port)
	(cond ((yaml-multiline? val)
	       (newline port)
	       (with-indent 2 (yaml-dump val port)))
	      (else
	       (yaml-dump val port)
	       (newline port))))))))


;;; Scalar types
;; !!bool
(define-method yaml-dump ((bool <boolean>) port)
  (display (if bool "true" "false") port))

(define (nan? x)
  (= x #i0/0))

(define (infinite? x)
  (or (= x #i1/0)
      (= x #i-1/0)))

;; !!float and integer
(define-method yaml-dump ((num <real>) port)
  (cond ((and (negative? num) (infinite? num))
	 (display "-.inf" port))
	((infinite? num)
	 (display ".inf" port))
	((nan? num)
	 (display ".nan" port))
	(else
	 (display (number->string (exact->inexact num) 10) port))))

(define-method yaml-dump ((num <integer>) port)
  (display (number->string num 10) port))

;; !!null
(define-method yaml-dump ((undef <undefined-object>) port)
  (display "null" port))

;; !!timestamp
(define-method yaml-dump ((date <date>) port)
  (display
   (regexp-replace #/..$/ (date->string date "~Y-~m-~d ~H:~M:~S.~N ~z") ":\\0")
   port))

(define-method yaml-dump ((uvec <u8vector>) port)
  (yaml-dump-binary (u8vector->string uvec) port))

;; !!str
(define-method yaml-dump ((str <string>) port)
  (define (simple-string? str)
    ;; http://yaml.org/type/null.html
    (define (yaml-null? s)
      (#/^null$/i s))
    ;; http://yaml.org/type/bool.html
    (define (bool? s)
      (#/^(?:y|yes|n|no|true|false|on|off)$/i s))
    ;; http://yaml.org/type/int.html
    (define (int? s)
      (#/^[-+]?(?:0|0b[0-1_,]+|0[0-7_,]+|[1-9][0-9_,]*|0x[0-9a-fA-F_]+|[1-9][0-9_,]*(?::[0-5]?[0-9])+)$/ s))
    ;; http://yaml.org/type/float.html
    (define (float? s)
      (or (#/[+-]?\.inf|\.nan/i s)
	  (#/^[-+]?(?:(?:[0-9][0-9_,]*)?\.[0-9.]*(?:[eE][-+][0-9]+)?|[0-9][0-9_,]*(?::[0-5]?[0-9])+\.[0-9_]*)$/ s)))
    ;; http://yaml.org/type/timestamp.html
    (define (timestamp? s)
      (#/^(?:\d{4}-\d{2}-\d{2}|\d{4}-\d\d?-\d\d?(?:[Tt]|[ \t]+)\d\d?:\d{2}:\d{2}(?:\.\d{2}*)?[ \t]*(?:Z|[-+]\d\d?(?::\d\d)?)?)$/ s))
    ;; http://yaml.org/spec/current.html#id2535812
    (define (plain? s)
      (#/^([$()+.\/0-9\;<=?A-Z\\^_`a-z~]|[-?:][!\"$%&'()*+,-.\/0-9\;<=>?@A-Z\[\\\]^_`a-z{|}~])([ !\"$%&'()*+,-.\/0-9\;<=>?@A-Z\[\\\]^_`a-z{|}~]|[!\"$%&'()*+,-.\/0-9\;<=>?@A-Z\[\\\]^_`a-z{|}~]\#|:[!\"$%&'()*+,-.\/0-9\;<=>?@A-Z\[\\\]^_`a-z{|}~])*$/ str))
    (and (not (yaml-null? str))
	 (not (bool? str))
	 (not (int? str))
	 (not (float? str))
	 (not (timestamp? str))
	 (plain? str)
	 #t))
  (cond ((string-incomplete? str)
	 (yaml-dump-binary str port))
	((simple-string? str)
	 (display str port))
	(else
	 (write str port))))

(define (yaml-dump-binary str port)
  (let1 s (base64-encode-string str)
    (display (indent "!binary |") port)
    (newline port)
    (with-indent 2
      (call-with-input-string (base64-encode-string str) 
	(lambda (in)
	  (port-for-each
	   (lambda (line)
	     (display (indent line) port)
	     (newline port))
	   (cut read-line in)))))))

#|
(use srfi-14)
(use util.sre)

(define (plain? str)
  (let* ((b-char      #[\r\n])
	 (c-indicator #[-?:,\[\]{}#&*!|>'\"%@`])
	 (c-printable #[\t\r\n !\"#$%&'()*+,-./0-9:\;<=>?@A-Z\[\\\]^_`a-z{|}~])
	 (nb-char             (char-set-difference c-printable b-char))
	 (nb-plain-char       (char-set-difference nb-char #[\t#:]))
	 (ns-plain-char       (char-set-difference nb-plain-char #[ ]))
	 (ns-plain-first-char (char-set-difference ns-plain-char c-indicator)))
    (and ((rx (seq bos
		   (or ,ns-plain-first-char
		       (seq ("-?:") ,ns-plain-char))
		   (* (or ,nb-plain-char
			  (seq ,ns-plain-char "#")
			  (seq ":" ,ns-plain-char)))
		   eos))
	  str)
	 #t)))
|#

(provide "text/yaml")
