/*
 * yaml.c
 *
 * Copyright (c) 2006 OOHASHI Daichi, All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <syck.h>

#include "yaml.h"
#include <gauche/uvector.h>

#define SCM_LIST8(a,b,c,d,e,f,g,h)					\
  Scm_Cons(a, Scm_Cons(b, Scm_Cons(c, (SCM_LIST5(d, e, f, g, h)))))

static const unsigned int HASH_INIT_SIZE = 16;
static ScmObj sym_merge = SCM_NIL;
static ScmObj sym_default = SCM_NIL;
static ScmObj make_date_proc = SCM_NIL;


static void
merge_hash_table(ScmObj ht1, ScmObj ht2)
{
    ScmHashIter iter;
    ScmHashEntry *e;

    Scm_HashIterInit(SCM_HASH_TABLE(ht2), &iter);

    while ((e = Scm_HashIterNext(&iter)) != NULL) {
	Scm_HashTableAdd(SCM_HASH_TABLE(ht1), e->key, e->value);
    }
}

static ScmObj
gauche_mktime(int nanosec, int sec, int min, int hour, int day, int mon, int year, int offset)
{
    return Scm_Apply(make_date_proc,
		     SCM_LIST8(SCM_MAKE_INT(nanosec),
			       SCM_MAKE_INT(sec),
			       SCM_MAKE_INT(min),
			       SCM_MAKE_INT(hour),
			       SCM_MAKE_INT(day),
			       SCM_MAKE_INT(mon),
			       SCM_MAKE_INT(year),
			       SCM_MAKE_INT(offset)));
}

static void
str_blow_away_extra_dots(SyckNode *node)
{
    char *ptr, *end;

    ptr = node->data.str->ptr;
    end = ptr + node->data.str->len;

    while (ptr < end && *ptr != '.') {
	ptr++;
    }

    if (*ptr == '.')
	ptr++;

    for (; ptr < end; ptr++) {
	if (*ptr == '.') {
	    memmove(ptr, ptr + 1, end - ptr);
	    end--;
	}
    }

    node->data.str->len = end - node->data.str->ptr;
}

static ScmObj
syck_mktime(char *str, size_t len)
{
    char *ptr = str;
    char *end = str + len;
    long year, mon, day, hour, min, sec, nanosec, offset;
    ScmObj args;
    year = mon = day = hour = min = sec = nanosec = offset = 0;

    /* year */
    if (*ptr != '\0' && ptr < end) {
        year = strtol(ptr, &ptr, 10);
    }

    /* month*/
    if (*ptr != '\0' && ptr < end) {
        while (!isdigit(*ptr)) ptr++;

        mon = strtol(ptr, &ptr, 10);
    }

    /* day */
    if (*ptr != '\0' && ptr < end) {
        while (!isdigit(*ptr)) ptr++;

        day = strtol(ptr, &ptr, 10);
    }

    /* hour */
    if (*ptr != '\0' && ptr < end) {
        while (!isdigit(*ptr)) ptr++;

        hour = strtol(ptr, &ptr, 10);
    }

    /* minute */
    if (*ptr != '\0' && ptr < end) {
        while (!isdigit(*ptr)) ptr++;

        min = strtol(ptr, &ptr, 10);
    }

    /* second */
    if (*ptr != '\0' && ptr < end) {
        while (!isdigit(*ptr)) ptr++;

        sec = strtol(ptr, &ptr, 10);
    }

    /* fraction */
    if (ptr < end && *ptr == '.') {
	ptr++;
	char buf[] = "000000000"; /* with nanosecond precision. */
	char *p = ptr, *q = buf;

	while (isdigit(*p) && p - ptr < 9) {
	    *q = *p;
	    q++;
	    p++;
	}

        nanosec = strtol(buf, NULL, 10);
    } else {
        nanosec = 0;
    }

    /* timezone */
    while (ptr < end &&
	   *ptr != 'Z' && *ptr != '+' && *ptr != '-' && *ptr != '\0')
      ptr++;

    if (ptr < end && (*ptr == '-' || *ptr == '+')) {
        offset = strtol(ptr, NULL, 10) * 3600;

        while (*ptr != ':' && *ptr != '\0') ptr++;

        if (*ptr == ':') {
            ptr++;

            if (offset < 0) {
                offset -= strtol(ptr, NULL, 10) * 60;
            } else {
                offset += strtol(ptr, NULL, 10) * 60;
            }
        }
    }
    
    return gauche_mktime(nanosec, sec, min, hour, day, mon, year, offset);
} 


SYMID
gauche_syck_parser_handler(SyckParser *parser, SyckNode *node)
{
    ScmObj obj;
    SYMID oid;
    int i;

    switch (node->kind) {
    case syck_str_kind:
	if (node->type_id == NULL || strcmp(node->type_id, "str") == 0) {
	    obj = SCM_MAKE_STR_COPYING(node->data.str->ptr);
	} else if (strcmp(node->type_id, "null") == 0) {
	    obj = SCM_UNDEFINED;
	} else if (strcmp(node->type_id, "bool#yes") == 0) {
	    obj = SCM_TRUE;
	} else if (strcmp(node->type_id, "bool#no") == 0) {
	    obj = SCM_FALSE;
	} else if (strcmp(node->type_id, "int#oct") == 0) {
	    long ival;
	    syck_str_blow_away_commas(node);
	    ival = strtol(node->data.str->ptr, NULL, 8);

	    obj = SCM_MAKE_INT(ival);
	} else if (strcmp(node->type_id, "int") == 0) {
	    long ival;
	    syck_str_blow_away_commas(node);
	    ival = strtol(node->data.str->ptr, NULL, 10);

	    obj = SCM_MAKE_INT(ival);
	} else if (strcmp(node->type_id, "int#hex") == 0) {
	    long ival;
	    syck_str_blow_away_commas(node);
	    ival = strtol(node->data.str->ptr, NULL, 16);

	    obj = SCM_MAKE_INT(ival);
	} else if (strcmp(node->type_id, "int#base60") == 0) {
	    char *ptr = node->data.str->ptr;
	    char *end = node->data.str->ptr + node->data.str->len;
	    long sixty = 1;
	    long total = 0;

	    syck_str_blow_away_commas(node);

	    while (end > ptr)
	    {
		long bnum = 0;
		char *colon = end - 1;

		while (colon >= ptr && *colon != ':') {
		    colon--;
		}

		if (colon > ptr && *colon == ':')
		    *colon = '\0';

		bnum = strtol(colon + 1, NULL, 10);
		total += bnum * sixty;
		sixty *= 60;
		end = colon;
	    }

	    obj = SCM_MAKE_INT(total);
	} else if (strcmp(node->type_id, "float#nan") == 0) {
	    obj = Scm_MakeFlonum(0.0 / 0.0);
	} else if (strcmp(node->type_id, "float#inf") == 0) {
	    obj = Scm_MakeFlonum(1.0 / 0.0);
	} else if (strcmp(node->type_id, "float#neginf") == 0) {
	    obj = Scm_MakeFlonum(-1.0 / 0.0);
	} else if (strcmp(node->type_id, "float#fix") == 0 ||
	           strcmp(node->type_id, "float#exp") == 0) {
	    double dval;
	    syck_str_blow_away_commas(node);
	    str_blow_away_extra_dots(node);
	    dval = strtod(node->data.str->ptr, NULL);

	    obj = Scm_MakeFlonum(dval);
	} else if (strcmp(node->type_id, "float#base60") == 0) {
	    char *ptr = node->data.str->ptr;
	    char *end = node->data.str->ptr + node->data.str->len;
	    long sixty = 1;
	    double total = 0.0;

	    syck_str_blow_away_commas(node);

	    while (end > ptr) {
		double bnum = 0;
		char *colon = end - 1;

		while (colon >= ptr && *colon != ':') {
		    colon--;
		}

		if (colon > ptr && *colon == ':')
		    *colon = '\0';

		bnum = strtod(colon + 1, NULL);
		total += bnum * sixty;
		sixty *= 60;
		end = colon;
	    }

	    obj = Scm_MakeFlonum(total);
	} else if (strcmp(node->type_id, "timestamp") == 0 ||
		   strcmp(node->type_id, "timestamp#spaced") == 0 ||
		   strcmp(node->type_id, "timestamp#iso8601") == 0) {
	    obj = syck_mktime(node->data.str->ptr, node->data.str->len);
	} else if (strcmp(node->type_id, "timestamp#ymd") == 0) {
	    long year = 0, month = 0, day = 0;
	    char *ptr = node->data.str->ptr;
	    char *end = ptr + node->data.str->len;

	    year = strtol(ptr, &ptr, 10);
	    while (*ptr && !isdigit(*ptr)) ptr++;

	    month = strtol(ptr, &ptr, 10);
	    while (*ptr && !isdigit(*ptr)) ptr++;

	    day = strtol(ptr, NULL, 10);

	    obj = gauche_mktime(0, 0, 0, 0, day, month, year, 0);
	} else if (strcmp(node->type_id, "binary") == 0) {
	    char *p, *q, *end, *buf, *dec;
	    long l = node->data.str->len;
	    p = node->data.str->ptr;
	    q = buf = syck_strndup(p, l);
	    end = p + l;

	    while (p < end) {
		if (*p != '\r' && *p != '\n' && *p != ' ') {
		    *q = *p;
		    q++;
		}
		p++;
	    }
	    *q = '\0';

	    dec = syck_base64dec(buf, q - buf);

            /* hmm... */
	    for (end = dec + (q - buf) - 1; end >= dec && *end != '\0';)
		end--;

	    obj = Scm_MakeU8VectorFromArray(end - dec, (unsigned char *)dec);

	    S_FREE(buf);
	    S_FREE(dec);
	} else if (strcmp(node->type_id, "merge") == 0) {
	    obj = sym_merge;
	} else if (strcmp(node->type_id, "default") == 0) {
	    obj = sym_default;
	} else {
	    obj = SCM_MAKE_STR_COPYING(node->data.str->ptr);
	}

	break;
    case syck_seq_kind:
	{
	    ScmObj elem, tail = SCM_NIL;
	    obj = tail;
	    
	    for (i = 0; i < node->data.list->idx; i++) {
		oid = syck_seq_read(node, i);
		syck_lookup_sym(parser, oid, (char **)&elem);

		SCM_APPEND1(obj, tail, elem);
	    }

	    break;
	}
    case syck_map_kind:
	{
	    ScmObj key, val;
	    obj = Scm_MakeHashTableSimple(SCM_HASH_EQUAL, HASH_INIT_SIZE);

	    for (i = 0; i < node->data.pairs->idx; i++) {
		oid = syck_map_read(node, map_key, i);
		syck_lookup_sym(parser, oid, (char **)&key);

		oid = syck_map_read(node, map_value, i);
		syck_lookup_sym(parser, oid, (char **)&val);
		
		if (Scm_EqP(key, sym_merge)) {
		    if (SCM_LISTP(val)) {
			SCM_FOR_EACH(val, val) {
			    merge_hash_table(obj, SCM_CAR(val));
			}
		    } else if (SCM_HASH_TABLE_P(val)) {
			merge_hash_table(obj, val);
		    } else {
			Scm_Error("internal error: bad value for !!merge.");
		    }
		} else if (Scm_EqP(key, sym_default)) {
		    continue; /* ignore */
		} else {
		    Scm_HashTablePut(SCM_HASH_TABLE(obj), key, val);
		}
	    }

	    break;
	}
    }

    oid = syck_add_sym(parser, (char *)obj);

    return oid;
}

static void
gauche_syck_error_handler(SyckParser *parser, char *msg)
{
    Scm_Error("syck error:%d: %s", parser->linect, msg);
}


ScmObj
yaml_parse_string(ScmString *str)
{
    SyckParser *parser;
    SYMID oid;
    ScmObj obj = SCM_EOF;
    int size = SCM_STRING_BODY_SIZE(SCM_STRING_BODY(str));

    parser = syck_new_parser();
    syck_parser_str(parser, Scm_GetString(str), size, NULL);
    syck_parser_handler(parser, gauche_syck_parser_handler);
    syck_parser_error_handler(parser, gauche_syck_error_handler);

    oid = syck_parse(parser);

    if (oid)
	syck_lookup_sym(parser, oid, (char **)&obj);

    syck_free_parser(parser);

    return obj;
}



/*
 * Module initialization function.
 */
extern void Scm_Init_yamllib(ScmModule*);

ScmObj Scm_Init_yaml(void)
{
    ScmModule *mod;

    /* Register this DSO to Gauche */
    SCM_INIT_EXTENSION(yaml);

    /* Create the module if it doesn't exist yet. */
    mod = SCM_MODULE(SCM_FIND_MODULE("text.yaml", TRUE));

    sym_merge = SCM_INTERN("merge");
    sym_default = SCM_INTERN("default");
    make_date_proc =
	Scm_EvalCString("make-date", SCM_OBJ(SCM_FIND_MODULE("srfi-19", 0)));

    /* Register stub-generated procedures */
    Scm_Init_yamllib(mod);
}
